//-----------------------------------------------------------------------------
//reductions.c
//Purpose: This program houses the functions needed to reduce an
//         assembly word and give resulting nesting index of word
//Author: Ryan Arredondo
//Date: 12.28.2011
//Update: 04.20.2012 - Fixed bug in sequences(w) function.
//						Possible that it could have gone out of subscript.
//					- Update to reduction algorithm since old one did not
//						recognize exact nesting index for all words
//		05.25.2012 - Translated to c
//----------------------------------------------------------------------------

//standard c libs
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>	//contains isdigit function
#include <math.h>	//contains pow function

//function templates
unsigned short ** step(unsigned short *, int, char, int *, int *);
int get_NI(unsigned short *, int, char);
unsigned short * getLetters(unsigned short *, int);
int * occurrences(unsigned short *, int, unsigned short);
short is_double_occurrence(unsigned short *, int);
unsigned short ** sequences(unsigned short *, int, int *);
unsigned short ** tau_sequences(unsigned short *, int, int *);
unsigned short ** sigma_sequences(unsigned short *, int, int *);
unsigned short * remove_seqs(unsigned short *, int, unsigned short **, int, int *);
short isInSeq(short, unsigned short **, int);
unsigned short * remove_ltr(unsigned short *, int, unsigned short);
unsigned short * relabel(unsigned short *, int);
void print_word(unsigned short *, int, short);
void file_print_word(FILE *, unsigned short *, int, short);
unsigned short * get_word(char *, int *);
void copy_words(unsigned short **, int *, unsigned short **, int *, int *);
void usage_message();
unsigned short * get_reverse(unsigned short *, int);
unsigned short ** get_isomorphisms(unsigned short *, int, int *);

////main function - called from reductions user interface
int main(int argc, char * argv[])
{
	unsigned short ** isomorphisms;
	unsigned short * word;
	char word_string[256];
	char type = 't';
	int counts[20];
	int NI = 0, size = 0, i = 0, bufferchar = 0, count = 0;
	FILE * InFile = NULL, * OutFile = NULL;
	
	for(i = 0; i < 20; i++) counts[i] = 0;
	for(i = 0; i < 256; i++) word_string[i] = (char) 0;
	
	if(argc < 2 || argc > 4){
		usage_message();
	}
	else if(argc == 2){	//input is direct word
		if(!(strncmp(argv[1], "-h", 2)) || !(strncmp(argv[1], "--help", 6)))
			usage_message();
		else{
			size = strlen(argv[1]);
			//converts chars to ints (shorts)
			for(i = 0; i < size; i++){
				if(!isdigit(argv[1][i]) && !ispunct(argv[1][i])){
					printf("Argument for word was not recognized \r\n");
					usage_message();
				}
			}
			word = get_word(argv[1], &size);
			
			NI = get_NI(word, size, type);
			print_word(word, size, 0);
			if(NI == -1) printf(": not DOW \r\n");
			else printf(": %d \r\n", NI);
			free(word);
			return 0;
		}
	}
	else if(argc == 3){
		if(!(strncmp(argv[1], "-t",2)) || !(strncmp(argv[1], "--text", 6)) || \
		!(strncmp(argv[1], "-c", 2)) || !(strncmp(argv[1], "--count", 7))){	
			//text file input
			InFile = fopen(argv[2], "r");
		}
		else if(!(strncmp(argv[1], "-i", 2)) || !(strncmp(argv[1], "--isomorphisms", 14))){
			size = strlen(argv[2]);
			//converts chars to ints (shorts)
			for(i = 0; i < size; i++){
				if(!isdigit(argv[2][i]) && !ispunct(argv[2][i])){
					printf("Argument for word was not recognized \r\n");
					usage_message();
				}
			}
			word = get_word(argv[2], &size);
			
			isomorphisms = get_isomorphisms(word, size, &count);	
			for(i = 0; i < count; i++){
				NI = get_NI(isomorphisms[i], size, type);
				print_word(isomorphisms[i], size, 0);
				printf(": %d\r\n", NI);
				free(isomorphisms[i]);
			}
			free(isomorphisms);
			free(word);
			return 0;
		}
		else{
			printf("Error interpreting input \r\n");
			usage_message();
		}
	}
	else if(argc == 4){
		if(!(strncmp(argv[1], "-t",2)) || !(strncmp(argv[1], "--text", 6))){
			InFile = fopen(argv[2], "r");
			OutFile = fopen(argv[3], "w");
		}
		else{
			printf("Error interpreting input \r\n");
			usage_message();
		}
	}
	
	if(InFile == NULL){
		printf("Couldn't open file: %s \r\n", argv[2]);
		exit(1);
	}
	if(argc == 4 && OutFile == NULL){
		printf("Couldn't open file: %s \r\n", argv[3]);
		exit(1);
	}
	
	//gets to first number in file
	do{
		bufferchar = fgetc(InFile);
	} while(!isdigit(bufferchar) && bufferchar != EOF);
	
	//goes til end of file
	while(bufferchar != EOF){
		size = 0;
		//gets word and puts to word_string
		while(!isspace(bufferchar) && bufferchar != EOF){
			word_string[size++] = (char) bufferchar;
			bufferchar = fgetc(InFile);
		}
		word = get_word(word_string, &size);
		NI = get_NI(word, size, type);
		
		//outputs word and nesting index
		if(NI != 0){
			if(!(strncmp(argv[1], "-c", 2)) || !(strncmp(argv[1], "--count", 7))){
				counts[NI - 1] += 1;
			}
			else{
				if(OutFile == NULL){	//print to console
					print_word(word, size, 0);
					printf(": %d\r\n", NI);
				}
				else{	//print to file
					file_print_word(OutFile, word, size, 0);
					fprintf(OutFile, ": %d\r\n", NI);
				}
			}
		}
		free(word);
		if(bufferchar != EOF) bufferchar = fgetc(InFile);
	}
	
	if(!(strncmp(argv[1], "-c", 2)) || !(strncmp(argv[1], "--count", 7))){
		for(i = 0; i < 20; i++){
			if(counts[i] != 0)
				printf("NI = %d: %d\r\n", i+1, counts[i]);
		}
	}
	
	fclose(InFile);
	if(OutFile != NULL) fclose(OutFile);
	return 0;
}

unsigned short * get_word(char * str_arg, int * size)
{
	int new_size = 0, i = 0, punct_ctr = 0;
	char * token = NULL;
	unsigned short * word = NULL;
	short isdelimited = 0;
	
	for(i = 0; i < *size && !isdelimited; i++)
		isdelimited = ispunct(str_arg[i]);

	if(!isdelimited){
		word = (unsigned short *) malloc(sizeof(unsigned short)*(*size));
		if(word == NULL){
			printf("mem could not be alloc'd for word(1)");
			exit(1);
		}
		//converts chars to ints
		for(i = 0; i < *size; i++)
			word[i] = (unsigned short) (str_arg[i] - '0');
	}	
	else{
		//counts delimiters so that correct amount of mem can be alloc'd
		for(i = 0; i < *size; i++)
			if(ispunct(str_arg[i])) punct_ctr++;
		word = (unsigned short *) calloc((punct_ctr + 1), sizeof(unsigned short));
		if(word == NULL){
			printf("mem could not be alloc'd for word(2)");
			exit(1);
		}
	
		new_size = 0;
		//word is written letter by letter by tokenizing each letter in str_arg
		token = strtok(str_arg, ",-.!#$%&'*+/");
		do{
			//parses token, makes conversion digit by digit
			for(i = 0; i < strlen(token); i++){
				word[new_size] += (unsigned short) \
				pow(10,(strlen(token) - i - 1))*(token[i] - '0');
			}
			new_size++;
			//sets up for next letter
			token = strtok(NULL, ",-.!#$%&'*+/");
		}while(token != NULL && new_size - 1 != punct_ctr);
		*size = new_size;
	}
	return word;
}

void usage_message(){
		printf("Try: \r\n\t./NestIndex 123321 or ./NestIndex 1,2,3,3,2,1 \r\n");
		printf("\t123321 and 1,2,3,3,2,1 can be replaced by any assembly word \r\n\r\n");
		printf("For text file input use: \r\n\t ./NestIndex -t Infile.txt \r\n\r\n");
		printf("For text file input with text output use: \r\n\t");
		printf("./NestIndex -t Infile.txt Outfile.txt\r\n\r\n");
		exit(0);
}

////////////////////////////////////////////////////////////////////////////////////////
//performs a step of reduction given an assembly word and a reduction type
//returns: words dobtained by removing all A and B type seqs from input word
//		or by removing any letters not in A or B type seq.
//		Returns NULL if step results in empty word.
unsigned short ** step(unsigned short * word, int size, char type, 
						int * count, int * sizes)
{
	unsigned short ** AB_list = NULL, ** return_list = NULL;
	unsigned short * letters = NULL;
	unsigned short drop_list[size/2];
	int seq_count = 0, drop_ctr = 0, i = 0, n = 0;
	int new_size = 0;	//variable gets size of first word in return_list
						//only important size since rest will be size - 2
	
	letters = getLetters(word, size);
	*count = 0;
	
	//if size <= 4, step results in empty word, regardless of assembly word given
	if(size <= 4){
		free(letters);
		return NULL;
	}
	
	if (type == 't') AB_list = tau_sequences(word, size, &seq_count);
	else AB_list = sigma_sequences(word, size, &seq_count);

	//Creates list of letters, not in AB seqs, to be dropped
	if(AB_list != NULL){	//AB_list is non-empty
		//checks if letters are in A or B type seq
		for(i = 0; i < size/2; i++){ //iterates over letters in word
		
			//if letters[i] is NOT in a seq, letter is added to drop_list
			if(!isInSeq(letters[i], AB_list, seq_count)) //push letter to drop list
				drop_list[drop_ctr++] = letters[i];
		}
		*count = drop_ctr + 1;	//=== # of seqs to be returned
	}
	else{ //every letter in word goes to drop list
		for(i = 0; i < size/2; i++)
			drop_list[i] = letters[i];
		*count = size/2;	//=== # of distinct letters in word
	}
	free(letters);
	
	//allocates mem for return_list
	return_list = (unsigned short **) \
		malloc(sizeof(unsigned short *)*(*count));
	if(return_list == NULL) {
		for(i = 0; i < seq_count; i++) free(AB_list[i]);
		free(AB_list);
		free(word);
		printf("mem could not be alloc'd for return_list");
		exit(1);
	}
	
	//return list gets word with seqs removed in first position
	if(AB_list != NULL){
		return_list[0] = remove_seqs(word, size, AB_list, seq_count, &new_size);
		return_list[0] = relabel(return_list[0], new_size);
		if(new_size == 0){
			//frees allocated memory
			free(return_list[0]);
			free(return_list);
			for(i = 0; i < seq_count; i++) free(AB_list[i]);
			free(AB_list);
			return NULL;
		}
		sizes[0] = new_size;
		
		//frees memory allocated for seqs
		for(i = 0; i < seq_count; i++)
			free(AB_list[i]);
		free(AB_list);
	}

	//this gets words with letter from drop_list removed
	n = 0;
	for(i = (AB_list == NULL)? 0: 1; i < *count; i++){
		return_list[i] = remove_ltr(word, size, drop_list[n++]);
		return_list[i] = relabel(return_list[i], size - 2);
		sizes[i] = size - 2;
	}
	return return_list;
}

////get_NI function
//function returns nesting index of a word
int get_NI(unsigned short * word, int size, char type)
{
	unsigned short ** current_words = NULL, ** next_words = NULL, ** step_words = NULL;
	int * step_sizes = NULL, * next_sizes = NULL, * current_sizes = NULL;
	int sizes_sum = 0;
	int step_count = 0, current_count = 0, next_count = 0;
	int i = 0, j = 0, NI = 0;
		
	//checks if word is double occurrence
    if(!is_double_occurrence(word, size)) return -1;

	//if word is empty word
	NI = 0;
	if(size == 0) return NI;
	
	step_sizes = (int *) malloc(sizeof(int)*(size/2));
	if(step_sizes == NULL) {
		free(word);
		printf("mem could not be alloc'd for step_sizes");
		exit(1);
	}
	current_count = 0;
	step_words = step(word, size, type, &current_count, step_sizes);
	NI++;
	if(step_words == NULL){	//first step gives empty word
		free(step_sizes);
		return NI;
	}

	current_words = (unsigned short **) malloc(sizeof(unsigned short *)*(current_count));
	current_sizes = (int *) malloc(sizeof(int)*(current_count));
	if(current_sizes == NULL || current_words == NULL){
		free(step_sizes);
		free(word);
		if(current_sizes != NULL) free(current_sizes);
		if(current_words != NULL) free(current_words);
		printf("mem could not be alloc'd for current_sizes/current_words");
		exit(1);
	}
	copy_words(current_words, current_sizes, step_words, step_sizes, &current_count);
	free(step_words);
	step_words = NULL;

	//checks if any current_word is empty
	while(1){
		NI++;
		//preparation for the next allocation of memory
		sizes_sum = 0;
		for(i = 0; i < current_count; i++)
			sizes_sum += current_sizes[i];
		sizes_sum /= 2;	//an upper bound on # of branch points from current words

		/*
		//for stats
		printf("sizes_sum is %d\r\n", sizes_sum);
		printf("current_count is %d\r\n", current_count);
		//for stats
		*/

		//mem allocated using upper bound obtained above
		next_words = malloc(sizeof(unsigned short *)*sizes_sum);
		next_sizes = malloc(sizeof(int)*sizes_sum);
		if(next_sizes == NULL || next_words == NULL){
			free(step_sizes);
			free(step_words);
			for(i = 0; i < current_count; i++) free(current_words[i]);
			free(current_words);
			if(next_sizes != NULL) free(next_sizes);
			if(next_words != NULL) free(next_words);
			free(word);
			printf("mem could not be alloc'd for next_sizes/next_words");
			exit(1);
		}
		next_count = 0;
		for(i = 0; i < current_count; i++){ //iterates over words in current_words
			step_count = 0;
			step_words = step(current_words[i], current_sizes[i], 
						type, &step_count, step_sizes);
			if(step_words == NULL){
				for(j = 0; j < current_count; j++) free(current_words[j]);
				free(current_words);
				free(current_sizes);
				for(j = 0; j < next_count; j++) free(next_words[j]);
				free(next_words);
				free(next_sizes);
				free(step_sizes);
				return NI;
			}
			
			memcpy((next_words + next_count), step_words, 
					sizeof(unsigned short *)*(step_count));
			memcpy((next_sizes + next_count), step_sizes,
					sizeof(int)*(step_count));
			next_count += step_count;

			//contents of step_words not freed because
			//contents (addresses) were just copied to next_words
			//hence freeing them would cause TROUBLE!!
			free(step_words);
			step_words = NULL;
		}
		for(i = 0; i < current_count; i++) free(current_words[i]);
		free(current_words);
		current_words = NULL;
		free(current_sizes);
		current_sizes = NULL;

		////Step complete
		//Current words become next_words
		current_words = (unsigned short **) \
			malloc(sizeof(unsigned short *)*(next_count));
		current_sizes = (int *) malloc(sizeof(int)*(next_count));
		if(current_words == NULL || current_sizes == NULL){
			free(step_words);
			free(step_sizes);
			free(next_sizes);
			for(i = 0; i < next_count; i++) free(next_words[i]);
			if(current_words != NULL) free(current_words);
			if(current_sizes != NULL) free(current_sizes);
			free(word);
			printf("mem could not be alloc'd for current_words/current_sizes");
			exit(1);
		}
		copy_words(current_words, current_sizes, next_words, next_sizes, &next_count);
		current_count = next_count;

		//contents of next_words not freed because
		//contents (addresses) were just copied to current_words
		//hence freeing them would cause TROUBLE!!
		free(next_words);
		free(next_sizes);
	}
}

void copy_words(unsigned short ** dest, int * dest_sizes,
	unsigned short ** source, int * source_sizes, int * count)
{
	int i = 0, j = 0, ctr = 0;
	short isInDest = 0;
	
	for(i = 0; i < *count; i++){
		isInDest = 0;
		for(j = 0; j < ctr && !(isInDest); j++){
			if(source_sizes[i] == dest_sizes[j]){	//could be a match
				isInDest = \
					(memcmp(source[i], dest[j], sizeof(short)*(dest_sizes[j])) == 0);
			}
		}
		if(!isInDest){	//add seq to dest
			j *= 2;
			memcpy((dest + ctr), (source + i), sizeof(short *));
			dest_sizes[ctr] = source_sizes[i];
			ctr++;
		}
		else free(source[i]);
	}
	*count = ctr;
}

////////////////////////////////////////////////////////////

////getLetters function
//argument passed:
//	letters array, word array, size of word
//	-NOTE: issues may occur if word is not double occurrence
//returns: no return
unsigned short * getLetters(unsigned short * word, int size)
{
	unsigned short * letters = calloc(size, sizeof(unsigned short));
	short notInAlphabet = 0;
	short letter_ctr = 0;
	int i = 0, j =0;
	
	if(letters == NULL){
		free(word);
		printf("mem could not be alloc'd for letters");
		exit(1);
	}
    	
	for(i = 0; i < size; i++){
		notInAlphabet = 1;	//evaluates true

		for(j = 0; j < letter_ctr; j++){
	    		if(word[i] == letters[j]) notInAlphabet = 0;	//evaluates false
		}
		if(notInAlphabet)
			letters[letter_ctr++] = word[i];
	}
	return letters;
}

////occurrences function
//Arguments passed: A word w, size of w, a letter in w, and pointer to occs
//Returns: none
int * occurrences(unsigned short * word, int size, unsigned short letter)
{
	int * occs = (int *) malloc(sizeof(int)*2);
	int n = 0, i = 0;
	
    	for(i = 0; i < size; i++){
		if(word[i] == letter){
			occs[n++] = i;
		}			
	}
	return occs;
}

////is double occurrence function
//arguments passed: a word w
//returns: 1 if w is double occurrence and 0 otherwise
short is_double_occurrence(unsigned short * word, int size)
{
	int i = 0, j = 0;
	short ltr_ctr = 0;
	unsigned short * letters;
	
    if((size%2) == 1){  //Word is not of even length
		return 0;
    }
    else{
		letters = getLetters(word, size);
	
		for(i = 0; i < size/2; i++){	//iterates over ltr in letters
			ltr_ctr = 0;
			//iterates over short in word and gets count of ltr in word
			for(j = 0; j < size; j++){
				if (letters[i] == word[j]) ltr_ctr++;
			}
			if(ltr_ctr != 2){
				free(letters);
				return 0;
			}
		}
		free(letters);
		return 1;
    }
}

////sequences function
//argument passed: a double occurence word w
//returns: type A and type B sequences in w
unsigned short ** sequences(unsigned short * word, int size, int * seq_ctr)
{
	/*Return var (seqs): array of arrays, holds sequences found*/
    unsigned short ** seqs = (unsigned short **) \
		malloc(sizeof(unsigned short *)*(size/2));
    unsigned short * letters;
	int i = 0, j = 0, n = 0;
	int * occs;
	
	if(seqs == NULL){
		free(word);
		printf("mem could not be alloc'd for seqs");
		exit(1);
	}
	*seq_ctr = 0;
	letters = getLetters(word, size);

    	//loops over the letters of w
    	for(i = 0; i < size/2; i++){
		n = 0;
		occs = occurrences(word, size, letters[i]);	//gets occs


		//checks for seqs of type A
		if ((occs[1] - occs[0])%2 != 0){
			while(word[occs[0] + n] == word[occs[1] - n] && \
				((occs[1] - n) - (occs[0] + n)) > 0){
				if(((occs[1] - n) - (occs[0] + n)) == 1){ 	//A type seq found					
					//allocates memory for seq to be stored
					seqs[*seq_ctr] = (unsigned short *) \
						malloc(sizeof(unsigned short)*(occs[1] - occs[0] + 1));
					//check that mem was allocated okay
					if(seqs[*seq_ctr] == NULL){
						for(i = 0; i < *seq_ctr - 1; i++)
							free(seqs[i]);
						free(seqs);
						free(word);
						printf("mem could not be alloc'd for seqs[*seq_ctr](1)");
						exit(1);
					}
					
					for(j = occs[0]; j <= occs[1]; j++){
						//Puts seq into seqs[seq_ctr - 1] letter by letter
						seqs[*seq_ctr][j - occs[0]] = word[j];
					}
					*seq_ctr += 1;
				}
				n++;
			}
		}
		//checks for seqs of type B
		n = 1; 	//to prevent a loop from being double counted as A and B
    	if((occs[1] + (occs[1] - occs[0] - 1)) < size && size != 2){   
			while(occs[1] + n < size && \
				word[occs[0] + n] == word[occs[1] + n] && \
				((occs[1] - (occs[0] + n))) > 0){
				if((occs[1] - (occs[0] + n)) == 1)		//B type seq found
				{
					//allocate mem for seq
					seqs[*seq_ctr] = (unsigned short *) \
						malloc(sizeof(unsigned short)*((occs[1] - occs[0])*2));
						
					//check that mem was allocated okay

					if(seqs[*seq_ctr] == NULL){
						for(i = 0; i < *seq_ctr - 1; i++)
							free(seqs[i]);
						free(seqs);
						free(word);
						printf("mem could not be alloc'd for seqs[*seq_ctr](2)");
						exit(1);
					}

					for(j = occs[0]; j <= (occs[1]+n); j++){
						//Puts seq into seqs[seq_ctr - 1] letter by letter
						seqs[*seq_ctr][j - occs[0]] = word[j];
					}
					*seq_ctr += 1;
				}
				n++;
			}
		}
    	free(occs);
    	}
	free(letters);
    return seqs;
}

////tau_sequences function
//arg passed: an assembly word w
//returns maximal subwords of w using sequences(w)
unsigned short ** tau_sequences(unsigned short * word, int size, int * tau_ctr)
{
	unsigned short ** tau_seqs = (unsigned short **) \
		malloc(sizeof(unsigned short *)*(size/2));
	short diff[size - 1];
	int i = 0, ctr = 0;
	short state = 0, place = 0;
	
	if(tau_seqs == NULL){
		free(word);
		printf("mem could not be alloc'd for tau_seqs");
		exit(1);
	}
	*tau_ctr = 0;
	
	//populates diff_word
	for(i = 0; i < size - 1; i++){
		diff[i] = (word[i+1] - word[i]);
	}
	
	//checks for A type seqs
	state = 0;
	place = 0;
	ctr = 0;
	for(i = 0; i < size - 1; i++){
		switch(diff[i])
		{
			case 0:
				if(state == 1 && i != size - 2) state = 2;	//expecting -1;
				else{	//must be a loop
					//allocate space for loop
					tau_seqs[*tau_ctr] = (unsigned short *) malloc(sizeof(short)*2);
					if(tau_seqs[*tau_ctr] == NULL){	
						for(i = 0; i < *tau_ctr; i++) free(tau_seqs[i]);
						free(tau_seqs);
						free(word);
						printf("mem could not be alloc'd for loop");
						exit(1);
					}
					memcpy(tau_seqs[*tau_ctr], (word + i), sizeof(short)*2);
					*tau_ctr += 1;
					state = 0;
					ctr = 0;
				}
				break;
			case 1:
				if(state == 0) place = i;	//holds placement of first 1
				else if(state == 2){
					place += ctr;
					tau_seqs[*tau_ctr] = (unsigned short *) malloc(sizeof(short)*((i - place)+ 1));
					if(tau_seqs[*tau_ctr] == NULL){
						for(i = 0; i < *tau_ctr; i++) free(tau_seqs[i]);
						free(tau_seqs);
						free(word);
						printf("mem could not be alloc'd for A type seq (1)");
						exit(1);
					}
					memcpy(tau_seqs[*tau_ctr], (word + place), sizeof(short)*((i - place) + 1));
					*tau_ctr += 1;
					place = i;
					ctr = 0;
				}					
				state = 1;	//expecting 0;
				ctr++;
				break;
			case -1:
				if(state == 2){
					ctr--;
					if(ctr == 0 || (i == size - 2)){	//a type seq found
						if(i == size - 2)	place += ctr;
						tau_seqs[*tau_ctr] = (unsigned short *) \
							malloc(sizeof(short)*((i - place) + 2));
						if(tau_seqs[*tau_ctr] == NULL){	
							for(i = 0; i < *tau_ctr; i++) free(tau_seqs[i]);
							free(tau_seqs);
							free(word);
							printf("mem could not be alloc'd for loop");
							exit(1);
						}
						memcpy(tau_seqs[*tau_ctr], (word + place), \
							sizeof(short)*((i - place) + 2));
						*tau_ctr += 1;
						state = 0;
						ctr = 0;
					}
				}
				else{
					state = 0;
					ctr = 0;
				}
				break;
			default:
				if(state == 2){	//a type seq is found
					place += ctr;
					tau_seqs[*tau_ctr] = (unsigned short *) \
						malloc(sizeof(short)*((i - place) + 1));
					if(tau_seqs[*tau_ctr] == NULL){	
						for(i = 0; i < *tau_ctr; i++) free(tau_seqs[i]);
						free(tau_seqs);
						free(word);
						printf("mem could not be alloc'd for loop");
						exit(1);
					}
					memcpy(tau_seqs[*tau_ctr], (word + place), \
						sizeof(short)*((i - place) + 1));
					*tau_ctr += 1;
				}		
				state = 0;
				ctr = 0;
		}
	}
	
	//checks for B type seqs
	state = 0;
	place = 0;
	ctr = 0;
	for(i = 0; i < size - 1; i++){
		switch(diff[i]){
			case 1:
				if(state == 0){
					state = 1;
					place = i;
					ctr++;
				}
				else if(state == 1){
					ctr++;
				}
				else if(state == 2){
					ctr--;
					if(ctr == 0){	//B type seq found
						tau_seqs[*tau_ctr] = (unsigned short *) \
							malloc(sizeof(short)*((i - place) + 2));
						if(tau_seqs[*tau_ctr] == NULL){
							for(i = 0; i < *tau_ctr; i++) free(tau_seqs[i]);
							free(tau_seqs);
							free(word);
							printf("mem could not be alloc'd for B type seq");
							exit(1);
						}
						memcpy(tau_seqs[*tau_ctr], (word + place), \
							sizeof(short)*((i - place) + 2));
						*tau_ctr += 1;
						state = 0;
						ctr = 0;					
					}
				}
				break;
			default:
				if(state == 1){
					if(diff[i] < 0 && (-1)*ctr <= diff[i]){
						place += (ctr + diff[i]);
						ctr = (-1)*diff[i];	//number of 1s expected
						state = 2;
					}
					else{
						state = 0;
						ctr = 0;
					}
				}
				else{
					state = 0;
					ctr = 0;
				}
		}
	}
	
	if(*tau_ctr == 0){
		free(tau_seqs);
		return NULL;
	}
	/*
	//debug
	printf("Word is ");
	print_word(word, size, 1);
	printf("*tau_ctr returns %d\r\n", *tau_ctr);
	for(i = 0; i < *tau_ctr ; i++){
		int n = 1;
		while(tau_seqs[i][n] > tau_seqs[i][n-1]) n++;
		n *= 2;
		print_word(tau_seqs[i], n, 1);
	}
	//debug
	*/
	return tau_seqs;
}

////sigma_sequences function
//arg passed: an assembly word w
//returns maximal length subwords of w using sequences(w)
unsigned short ** sigma_sequences(unsigned short * word, int size, int * sig_ctr)
{
	int max_length = 0, seq_count = 0, i = 0, j = 0;
    unsigned short ** seqs = sequences(word, size, &seq_count);
	unsigned short ** sig_seqs = (unsigned short **)\
		malloc(sizeof(unsigned short *)*(seq_count));
	unsigned short * seq;
	short indices[seq_count]; //array encodes which seqs are greatest in length
	
	if (sig_seqs == NULL){
		for(i = 0; i < seq_count; i++) free(seqs[i]);
		free(seqs);
		free(word);
		printf("mem could not be alloc'd for sig_seqs");
		exit(1);
	}
	
	if(seq_count == 0){
		for(i = 0; i < seq_count; i++) free(seqs[i]);
		free(seqs);
		free(sig_seqs);
		return NULL;
	}
	*sig_ctr = 0;
	max_length = 0;
	
	for(i = 0; i < seq_count; i++){
		indices[i] = 1;
		seq = seqs[i];
		j = 1;
		//gets number of unique letters in seq and sets max_length accordingly
		while(seq[j] > seq[j-1]) j++;
		if(2*j > max_length){
			max_length = 2*j;
			for(j = 0; j < i; j++) indices[j] = 0;
		}
		else indices[i] = 0;
	}
	
    for(i = 0; i < seq_count; i++){
		if(indices[i]){	//seqs[i] is of maximal length
			sig_seqs[*sig_ctr] = (unsigned short *) \
				malloc(sizeof(unsigned short)*max_length);
			
			if(sig_seqs[*sig_ctr] == NULL){
				free(sig_seqs);
				for(i = 0; i < seq_count; i++) free(seqs[i]);
				free(seqs);
				free(word);
				printf("mem could not be alloc'd for sig_seqs[*sig_ctr]");
				exit(1);
			}

			memcpy(sig_seqs[*sig_ctr], seqs[i], sizeof(unsigned short)*max_length);
			*sig_ctr += 1;
		}
		free(seqs[i]);
		seqs[i] = NULL;
    }
	free(seqs);
    return sig_seqs;
}

////remove_sequences function
//args passed: an assembly word w and a set of a subwords of w
//returns the word obtained from the removal of subwords from w
unsigned short * remove_seqs(unsigned short * word, int size, 
			unsigned short ** seqs, int seq_count, int * new_size)
{
	unsigned short * new_word = (unsigned short *) \
		malloc(sizeof(unsigned short)*(size));
	int i = 0;
	
	if(new_word == NULL){
		free(word);
		printf("mem could not be alloc'd for new_word (remove_seqs)");
		exit(1);
	}
	*new_size = 0;

    	for(i = 0; i < size; i++){
		if(!isInSeq(word[i], seqs, seq_count)){
			new_word[*new_size] = word[i];
			*new_size += 1;
		}
    	}
    	return new_word;
}

short isInSeq(short letter, unsigned short ** seqs, int seq_count)
{
	unsigned short * seq;
	int i, j;

	for(i = 0; i < seq_count; i++){
		seq = seqs[i];
		j = 0;
		do{
			if(letter == seq[j++]) return 1;
		} while(seq[j] > seq[j-1]);
	}
	return 0;
}


////remove function
//args passed: an assembly word w and a letter of that assembly word
//returns word obtained from removing letter passed from word passed
unsigned short * remove_ltr(unsigned short * word, int size, unsigned short letter)
{
	unsigned short * new_word = (unsigned short *) \
		malloc(sizeof(unsigned short)*(size - 2));	//minus 2 for letter removed
	int i = 0, new_size = 0;
	
	if(new_word == NULL){
		free(word);
		printf("mem could not be alloc'd for new_word (remove_ltr)");	
		exit(1);
	}
	for(i = 0; i < size; i++){
		if(word[i] != letter)
			new_word[new_size++] = word[i];
	}
    return new_word;
}

////relabel function
//args passed: an assembly word w
//returns the word obtained from w after relabelling
unsigned short * relabel(unsigned short * word, int size)
{
	unsigned short * new_word = (unsigned short *) malloc(sizeof(unsigned short)*size);
	unsigned short new_ltr = 1;
	int i, j;
	
	if(new_word == NULL){
		free(word);
		printf("mem could not be alloc'd for new_word (relabel)");
		exit(1);
	}
	for(i = 0; i < size; i++){
		for(j = i+1; j < size; j++){
		    if (word[i] == word[j]){
			new_word[i] = new_ltr;
			new_word[j] = new_ltr++;
		    }
		}
	}
	free(word);
	return new_word;
}

void print_word(unsigned short * word, int size, short return_bool)
{
	int i = 0;
	if(size >= 20){
		for(i = 0; i < size - 1; i++)
			printf("%u,", word[i]);
		printf("%u", word[size - 1]);
	}
	else{
		for(i = 0; i < size; i++)
			printf("%u", word[i]);
	}
	if(return_bool)	printf("\r\n");
}

void file_print_word(FILE * file, unsigned short * word, int size, short return_bool)
{
	int i = 0;
	if(size >= 20){
		for(i = 0; i < size - 1; i++)
			fprintf(file, "%u,", word[i]);
		fprintf(file, "%u", word[size - 1]);
	}
	else{
		for(i = 0; i < size; i++)
			fprintf(file, "%u", word[i]);
	}
	if(return_bool)	fprintf(file, "\r\n");
}

unsigned short * get_reverse(unsigned short * word, int size)
{
	unsigned short * reverse = (unsigned short *) malloc(sizeof(short)*(size));
	int i = 0;

	if(reverse == NULL){
		free(word);
		printf("mem could not be alloc'd for reverse");
		exit(1);
	}
	
	for(i = 0; i < size; i++)
		reverse[i] = word[(size - 1) - i];
	
	reverse = relabel(reverse, size);
	return reverse;
}

unsigned short ** get_isomorphisms(unsigned short * word, int size, int * count){
	unsigned short ** isomorphisms;
	unsigned short * temp_word, * reverse;
	int i = 0, offset;
	short isDup = 0;
	
	*count = 0;
	isomorphisms = (unsigned short **) malloc(sizeof(short *)*(size));
	temp_word = (unsigned short *) malloc(sizeof(short)*size);
	
	if(isomorphisms == NULL || temp_word == NULL){
		free(word);
		if(isomorphisms != NULL) free(isomorphisms);
		if(temp_word != NULL) free(temp_word);
		printf("mem could not be alloc'd for isomorphisms");
		exit(1);
	}

	offset = 0;
	isDup = 0;
	while(offset < size && !isDup){
		for(i = 0; i + offset < size; i++)
			temp_word[i] = word[i + offset];
		
		for(i = 0; i < offset; i++)
			temp_word[(size - offset) + i] = word[i];
		
		temp_word = relabel(temp_word, size);
		
		isDup = 0;
		for(i = 0; i < *count && !isDup; i++)
			isDup = (memcmp(isomorphisms[i], temp_word, sizeof(short)*size) == 0);
		if(!isDup){
			isomorphisms[*count] = temp_word;
			*count += 1;
			temp_word = malloc(sizeof(short)*(size));
			if(temp_word == NULL){
				for(i = 0; i < *count; i++) free(isomorphisms[i]);
				free(isomorphisms);
				free(word);
				printf("mem could not be alloc'd for temp_word(1)");
				exit(1);
			}
			offset++;
		}
	}
	
	//same as above, only with reverse.
	reverse = get_reverse(word, size);
	offset = 0;
	isDup = 0;
	while(offset < size && !isDup){
		for(i = 0; i + offset < size; i++)
			temp_word[i] = reverse[i + offset];
		
		for(i = 0; i < offset; i++)
			temp_word[(size - offset) + i] = reverse[i];
		
		temp_word = relabel(temp_word, size);
		
		isDup = 0;
		for(i = 0; i < *count && !isDup; i++)
			isDup = (memcmp(isomorphisms[i], temp_word, sizeof(short)*size) == 0);
		if(!isDup){
			isomorphisms[*count] = temp_word;
			*count += 1;
			temp_word = malloc(sizeof(short)*(size));
			if(temp_word == NULL){
				for(i = 0; i < *count; i++) free(isomorphisms[i]);
				free(isomorphisms);
				free(reverse);
				free(word);
				printf("mem could not be alloc'd for temp_word(2)");
				exit(1);
			}
			offset++;
		}
	}
	free(reverse);
	free(temp_word);	
	return isomorphisms;
}
		
